from typing import TYPE_CHECKING, Optional, overload

if TYPE_CHECKING:
    from earthscope_sdk.common.context import SdkContext
    from earthscope_sdk.config.settings import SdkSettings


class _SdkClient:
    @property
    def ctx(self):
        """
        SDK Context
        """
        return self._ctx

    @property
    def is_closed(self):
        """
        This client is closed
        """
        return self._ctx.is_closed

    @property
    def resources(self):
        """
        References to EarthScope resources
        """
        return self.ctx.settings.resources

    @overload
    def __init__(self): ...

    @overload
    def __init__(self, *, ctx: "SdkContext"): ...

    @overload
    def __init__(self, *, settings: "SdkSettings"): ...

    def __init__(
        self,
        *,
        ctx: Optional["SdkContext"] = None,
        settings: Optional["SdkSettings"] = None,
    ):
        # lazy imports
        from earthscope_sdk.common.context import SdkContext
        from earthscope_sdk.config.settings import SdkSettings

        if ctx:
            self._created_ctx = False

        else:
            if settings is None:
                settings = SdkSettings()

            ctx = SdkContext(settings=settings)
            self._created_ctx = True

        self._ctx = ctx


class AsyncSdkClient(_SdkClient):
    """
    Base class for asynchronous EarthScope SDK clients
    """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_t, exc_v, exc_tb):
        await self.close()

    async def close(self):
        """
        Close this client
        """
        # only close the context if we created it
        if self._created_ctx:
            await self.ctx.async_close()


class SdkClient(_SdkClient):
    """
    Base class for synchronous EarthScope SDK clients
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """
        Close this client
        """
        # only close the context if we created it
        if self._created_ctx:
            self.ctx.close()



from earthscope_sdk.common.service import SdkService


class UserBaseService(SdkService):
    """
    L1 user/IDM endpoints
    """

    async def _get_aws_credentials(self, *, role: str = "s3-miniseed"):
        """
        Retrieve temporary AWS credentials
        """

        from earthscope_sdk.client.user.models import AwsTemporaryCredentials

        req = self.ctx.httpx_client.build_request(
            method="GET",
            url=f"{self.resources.api_url}beta/user/credentials/aws/{role}",
        )

        resp = await self._send_with_retries(req)

        return AwsTemporaryCredentials.model_validate_json(resp.content)

    async def _get_profile(self):
        """
        Retrieve your EarthScope user profile
        """

        from earthscope_sdk.client.user.models import UserProfile

        req = self.ctx.httpx_client.build_request(
            method="GET",
            url=f"{self.resources.api_url}beta/user/profile",
        )

        resp = await self._send_with_retries(req)

        return UserProfile.model_validate_json(resp.content)



"""
This module facilitates bootstrapping SDK settings from a JSON-encoded environment variable.
"""

import json
import logging
import os

from pydantic_settings import PydanticBaseSettingsSource

logger = logging.getLogger(__name__)


class BootstrapEnvironmentSettingsSource(PydanticBaseSettingsSource):
    """
    This SettingsSource facilitates bootstrapping the SDK from a special environment variable.

    The environment variable should be a JSON string of the expected SDK settings and structure.
    """

    def __init__(self, settings_cls, env_var: str):
        super().__init__(settings_cls)
        self._env_var = env_var

    def __call__(self):
        try:
            bootstrap_settings = os.environ[self._env_var]
        except KeyError:
            return {}

        try:
            return json.loads(bootstrap_settings)
        except json.JSONDecodeError:
            logger.warning(
                f"Found bootstrap environment variable '{self._env_var}', but unable to decode content as JSON"
            )
            return {}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(env_var='{self._env_var}')"

    def get_field_value(self, *args, **kwargs): ...  # unused abstract method



from contextlib import suppress
from functools import cached_property
from typing import Annotated, Any, Type

from pydantic import Field, TypeAdapter, ValidationError, field_validator
from pydantic_settings import (
    BaseSettings,
    InitSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)

from earthscope_sdk.config._bootstrap import BootstrapEnvironmentSettingsSource
from earthscope_sdk.config._compat import LegacyEarthScopeCLISettingsSource
from earthscope_sdk.config._util import deep_merge, get_config_dir, slugify
from earthscope_sdk.config.error import ProfileDoesNotExistError
from earthscope_sdk.config.models import SdkBaseSettings, Tokens

_BOOTSTRAP_ENV_VAR = "ES_BOOTSTRAP_SETTINGS"
"""Environment variable for bootstrapping the SDK"""

_DEFAULT_PROFILE = "default"
"""Default profile name"""

_LOCAL_APP_DIR = get_config_dir(app_name="earthscope")
"""Local SDK application directory"""


def _get_config_toml_path():
    """
    Local SDK configuration file
    """
    return _LOCAL_APP_DIR / "config.toml"


def _get_profile_dir(profile_name: str):
    """
    Retrieve the local SDK directory for the named profile
    """
    return _LOCAL_APP_DIR / slugify(profile_name)


def _get_profile_tokens_file(profile_name: str):
    """
    Retrieve the local SDK tokens file for the named profile
    """
    profile_dir = _get_profile_dir(profile_name)
    return profile_dir / "tokens.json"


class _SdkGlobalSettings(BaseSettings):
    """
    EarthScope SDK global settings; not for direct use.

    This class only loads settings from config.toml.
    """

    defaults: Annotated[
        SdkBaseSettings,
        Field(validation_alias="default"),
    ] = {}

    profile_settings_by_name: Annotated[
        dict[str, SdkBaseSettings],
        Field(validation_alias="profile"),
    ] = {}

    model_config = SettingsConfigDict(
        extra="ignore",
        frozen=True,
        nested_model_default_partial_update=True,
        populate_by_name=True,
    )

    def get_merged(self, profile_name: str):
        """
        Get profile-specific settings merged with defaults
        """
        try:
            p = self.profile_settings_by_name[profile_name]
        except KeyError:
            raise ProfileDoesNotExistError(f"Profile '{profile_name}' does not exist")

        defaults = self.defaults.model_dump(
            exclude_none=True,
            exclude_unset=True,
            exclude_defaults=True,
        )
        profile_specific = p.model_dump(
            exclude_defaults=True,  # defer to top-level field defaults
            exclude_unset=True,  # only keep explicitly set fields
            exclude_none=True,
        )

        # order is imporant; profile-specific args override default args
        return deep_merge(defaults, profile_specific, profile_name=profile_name)

    @field_validator("profile_settings_by_name", mode="before")
    @classmethod
    def ensure_default_profile(cls, v: dict[str, dict]):
        """
        Ensure the default profile exists
        """
        v.setdefault(_DEFAULT_PROFILE, {})
        return v

    @classmethod
    def settings_customise_sources(cls, settings_cls: Type[BaseSettings], **_):
        """
        Override the loading chain to exclusively check ~/.earthscope/config.toml
        """
        # we get the file path dynamically to allow monkeypatching for tests
        toml_path = _get_config_toml_path()
        toml_settings = TomlConfigSettingsSource(settings_cls, toml_path)

        return (toml_settings,)


class _GlobalSettingsSource(PydanticBaseSettingsSource):
    """
    This SettingsSource facilitates falling back to settings configured in config.toml
    and state in tokens.json when loading the configuration chain in SdkSettings.

    Consolidates merged profile state with state loaded from a profile's tokens.json file
    """

    def __init__(self, settings_cls, *keys: str):
        super().__init__(settings_cls)
        self._keys = keys

    def __call__(self):
        # Extract profile
        profile_name = _DEFAULT_PROFILE
        for k in self._keys:
            with suppress(KeyError):
                profile_name = self._current_state[k]
                break

        # Load global merged settings
        merged = _SdkGlobalSettings().get_merged(profile_name)

        # Load from tokens.json
        tokens_json = {}
        with suppress(FileNotFoundError, ValidationError):
            raw = _get_profile_tokens_file(profile_name).read_bytes()
            tokens = Tokens.model_validate_json(raw)
            tokens_json = tokens.model_dump(
                exclude_none=True,
                exclude_unset=True,
                exclude_defaults=True,
            )

        # order is imporant; init args override tokens.json args
        return deep_merge({"oauth2": tokens_json}, merged)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def get_field_value(self, *args, **kwargs): ...  # unused abstract method


class _InitSettingsWithoutDefaultSource(InitSettingsSource):
    """
    InitSettingsSource behavior but excludes default values & unset fields
    to allow values from lower-precedence settings sources to bubble up.
    """

    def __init__(self, other: InitSettingsSource):
        super().__init__(
            other.settings_cls,
            other.init_kwargs,
            other.nested_model_default_partial_update,
        )
        self.dict_adapter = TypeAdapter(dict[str, Any])

    def __call__(self) -> dict[str, Any]:
        return self.dict_adapter.dump_python(
            self.init_kwargs,
            exclude_defaults=True,
            exclude_unset=True,
        )


class SdkSettings(SdkBaseSettings, BaseSettings):
    """
    EarthScope SDK settings.

    Use `profile_name` to utilize a named profile from `~/.earthscope/config.toml`

    Settings loading chain (order of precedence):
    1. arguments passed to constructor
    2. environment variables
    3. dotenv variables
    4. ~/.earthscope/config.toml
    """

    profile_name: Annotated[
        str,
        Field(validation_alias="es_profile"),
        # NOTE: aliases do not include the env_prefix by default; add it explicitly
    ]

    model_config = SettingsConfigDict(
        env_nested_delimiter="__",
        env_prefix="ES_",
        extra="ignore",
        frozen=True,
        nested_model_default_partial_update=True,
        populate_by_name=True,
    )

    @cached_property
    def profile_dir(self):
        """
        The path to the local SDK directory for the named profile
        """
        return _get_profile_dir(self.profile_name)

    @cached_property
    def tokens_file(self):
        """
        The path to the tokens file for the named profile
        """
        return _get_profile_tokens_file(self.profile_name)

    def delete_tokens(self, missing_ok=False):
        """
        Delete tokens from this named profile's state

        Args:
            missing_ok: Whether or not to throw an error if the file does not exist

        Returns:
            this auth flow
        """
        self.tokens_file.unlink(missing_ok=missing_ok)
        return self

    def write_tokens(self, tokens: Tokens):
        """
        Write tokens to this named profile's state

        Args:
            tokens: the tokens to persist in local storage
        """
        body = tokens.model_dump_json(
            include=Tokens.model_fields.keys(),
            indent=2,
            context="plaintext",  # write SecretStr's actual values in plaintext
            exclude_none=True,
        )
        self.tokens_file.parent.mkdir(parents=True, exist_ok=True)
        self.tokens_file.write_text(body)

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: InitSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        **_,
    ):
        """
        Override the loading chain to additionally check ~/.earthscope/config.toml
        """

        # Init settings, but without defaults & unset values
        init_settings = _InitSettingsWithoutDefaultSource(init_settings)

        # Check for all the ways profile name may be supplied
        alias = SdkSettings.model_fields["profile_name"].validation_alias
        global_settings = _GlobalSettingsSource(settings_cls, "profile_name", alias)

        # Check for bootstrapping configuration
        bootstrap_settings = BootstrapEnvironmentSettingsSource(
            settings_cls,
            _BOOTSTRAP_ENV_VAR,
        )

        # Compatibility with earthscope-cli v0.x.x state:
        # If we find this file, we only care about the access and refresh tokens
        keep_keys = {"access_token", "refresh_token"}
        legacy_settings = LegacyEarthScopeCLISettingsSource(settings_cls, *keep_keys)

        # Order of precedence
        return (
            init_settings,
            env_settings,
            dotenv_settings,
            global_settings,
            legacy_settings,
            bootstrap_settings,
        )



import json
from pathlib import Path
from textwrap import dedent

import pytest
from pydantic import ValidationError
from pytest import MonkeyPatch

from earthscope_sdk import __version__
from earthscope_sdk.config._compat import _get_legacy_auth_state_path
from earthscope_sdk.config.error import ProfileDoesNotExistError
from earthscope_sdk.config.models import (
    AuthFlowSettings,
    HttpRetrySettings,
    HttpSettings,
    Tokens,
)
from earthscope_sdk.config.settings import (
    _BOOTSTRAP_ENV_VAR,
    SdkSettings,
    _get_config_toml_path,
)

_bootstrap_state_json = json.dumps(
    {
        "oauth2": {
            "audience": "https://bootstrap-audience.earthscope.org",
            "client_id": "bootstrap-client-id",
            "domain": "https://bootstrap-domain.earthscope.org",
            "scope": "bootstrap-scope",
            "access_token": "bootstrap-at",
            "refresh_token": "bootstrap-rt",
            "id_token": "bootstrap-it",
        }
    }
)


@pytest.fixture
def config_toml():
    return _get_config_toml_path()


@pytest.fixture
def legacy_tokens_json():
    return _get_legacy_auth_state_path()


@pytest.fixture(scope="session")
def default_settings():
    return SdkSettings()


class TestSdkSettings:
    def test_defaults(self):
        s = SdkSettings()
        assert s.profile_name == "default"
        assert s.oauth2.audience == "https://api.earthscope.org"
        assert str(s.oauth2.domain) == "https://login.earthscope.org/"
        assert s.oauth2.client_id == "b9DtAFBd6QvMg761vI3YhYquNZbJX5G0"
        assert s.oauth2.scope == "offline_access"
        assert str(s.resources.api_url) == "https://api.earthscope.org/"
        assert s.http.user_agent == f"earthscope-sdk py/{__version__}"
        assert s.http.max_connections is None
        assert s.http.max_keepalive_connections is None
        assert s.http.keepalive_expiry.total_seconds() == 5.0
        assert s.http.timeout_read.total_seconds() == 5.0
        assert s.http.retry.attempts == 3
        assert s.http.retry.timeout.total_seconds() == 20.0
        assert s.http.retry.wait_initial.total_seconds() == 0.1
        assert s.http.retry.wait_max.total_seconds() == 5.0
        assert s.http.retry.wait_jitter.total_seconds() == 1.0
        assert s.http.retry.wait_exp_base == 2
        assert s.oauth2.client_secret is None
        assert s.oauth2.access_token is None
        assert s.oauth2.refresh_token is None
        assert s.oauth2.retry.attempts == 5
        assert s.oauth2.retry.timeout.total_seconds() == 30.0
        assert s.oauth2.retry.wait_initial.total_seconds() == 1.0
        assert s.oauth2.retry.wait_max.total_seconds() == 5.0
        assert s.oauth2.retry.wait_jitter.total_seconds() == 3.0
        assert s.oauth2.retry.wait_exp_base == 2

    def test_profile_does_not_exist_init(self):
        with pytest.raises(ProfileDoesNotExistError):
            SdkSettings(profile_name="foo")

    def test_profile_does_not_exist_env(self, monkeypatch: MonkeyPatch):
        monkeypatch.setenv("ES_PROFILE", "foo")
        with pytest.raises(ProfileDoesNotExistError):
            SdkSettings()

    def test_nested_init_with_dict(self):
        s = SdkSettings(oauth2={"scope": "dict-scope"})
        assert s.oauth2.scope == "dict-scope"

    def test_secret_serialization(self):
        s = SdkSettings(
            oauth2={
                "access_token": "foo-at",
                "id_token": "foo-it",
                "refresh_token": "foo-rt",
                "client_secret": "foo-secret",
            }
        )

        dumped = s.model_dump(mode="json")
        assert dumped["oauth2"]["access_token"] == "**********"
        assert dumped["oauth2"]["id_token"] == "**********"
        assert dumped["oauth2"]["refresh_token"] == "**********"
        assert dumped["oauth2"]["client_secret"] == "**********"

        dumped_plaintext = s.model_dump(mode="json", context="plaintext")
        assert dumped_plaintext["oauth2"]["access_token"] == "foo-at"
        assert dumped_plaintext["oauth2"]["id_token"] == "foo-it"
        assert dumped_plaintext["oauth2"]["refresh_token"] == "foo-rt"
        assert dumped_plaintext["oauth2"]["client_secret"] == "foo-secret"


class TestSdkSettingsPrecedence:
    def test_precedence_all(self, config_toml: Path, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.scope = "default-scope"

                [profile.default]
                oauth2.scope = "profile-scope"
            """)
        )
        monkeypatch.setenv("ES_OAUTH2__SCOPE", "env-scope")
        s = SdkSettings(oauth2=AuthFlowSettings(scope="init-scope"))

        assert s.oauth2.scope == "init-scope"

    def test_precedence_env(self, config_toml: Path, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.scope = "default-scope"

                [profile.default]
                oauth2.scope = "profile-scope"
            """)
        )
        monkeypatch.setenv("ES_OAUTH2__SCOPE", "env-scope")
        s = SdkSettings()

        assert s.oauth2.scope == "env-scope"

    def test_precedence_profile(self, config_toml: Path, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.scope = "default-scope"

                [profile.default]
                oauth2.scope = "profile-scope"
            """)
        )
        s = SdkSettings()

        assert s.oauth2.scope == "profile-scope"

    def test_precedence_defaults(self, config_toml: Path, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.scope = "default-scope"
            """)
        )
        s = SdkSettings()

        assert s.oauth2.scope == "default-scope"

    def test_precedence_bootstrap_state(self, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)

        s = SdkSettings()
        assert s.oauth2.audience == "https://bootstrap-audience.earthscope.org"
        assert s.oauth2.client_id == "bootstrap-client-id"
        assert str(s.oauth2.domain) == "https://bootstrap-domain.earthscope.org/"
        assert s.oauth2.scope == "bootstrap-scope"
        assert s.oauth2.access_token.get_secret_value() == "bootstrap-at"
        assert s.oauth2.refresh_token.get_secret_value() == "bootstrap-rt"
        assert s.oauth2.id_token.get_secret_value() == "bootstrap-it"


class TestSdkSettingsProfiles:
    def test_profile_settings(self, config_toml: Path):
        config_toml.write_text(
            dedent("""
                [profile.pytest]
                oauth2.scope = "pytest-scope"
            """)
        )
        s = SdkSettings(profile_name="pytest")

        assert s.oauth2.scope == "pytest-scope"

    def test_profile_merging(self, config_toml: Path, monkeypatch: MonkeyPatch):
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.scope = "default-scope"
                oauth2.audience = "default-audience"
                oauth2.client_id = "default-client-id"
                oauth2.domain = "http://default-domain"

                [profile.pytest]
                # intentionally omitting `oauth2.scope` to test merging
                oauth2.audience = "profile-audience"
                oauth2.client_id = "profile-client-id"
                oauth2.domain = "http://profile-domain"
            """)
        )

        # intentionally omitting scope and audience to test merging
        monkeypatch.setenv("ES_OAUTH2__CLIENT_ID", "env-client-id")
        monkeypatch.setenv("ES_OAUTH2__DOMAIN", "http://env-domain")

        # intentionally omitting scope, audience and domain to test merging
        s = SdkSettings(
            profile_name="pytest",
            oauth2=AuthFlowSettings(domain="http://init-domain"),
        )

        assert str(s.resources.api_url) == "https://api.earthscope.org/"
        assert s.oauth2.scope == "default-scope"
        assert s.oauth2.audience == "profile-audience"
        assert s.oauth2.client_id == "env-client-id"
        assert str(s.oauth2.domain) == "http://init-domain/"

    def test_multiple_profiles(self, config_toml: Path):
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.scope = "default-scope"
                oauth2.audience = "default-audience"
                oauth2.client_id = "default-client-id"
                oauth2.domain = "http://default-domain"

                [profile.foo]
                oauth2.scope = "foo-scope"

                [profile.bar]
                oauth2.scope = "bar-scope"
            """)
        )
        foo = SdkSettings(profile_name="foo")
        bar = SdkSettings(profile_name="bar")

        assert foo.oauth2.scope == "foo-scope"
        assert bar.oauth2.scope == "bar-scope"

    def test_legacy_state(self, legacy_tokens_json: Path):
        legacy_tokens_json.write_text(
            dedent("""
                {
                    "audience": "legacy-aud",
                    "domain": "https://legacy-domain.earthscope.org/",
                    "client_id": "legacy-client-id",
                    "scope": "legacy-scope",
                    "api_url": "https://legacy-api.earthscope.org/",
                    "user_agent": "legacy",
                    "refresh_token": "legacy-rt",
                    "access_token": "legacy-at"
                }
            """)
        )

        s = SdkSettings()

        assert s.oauth2.refresh_token.get_secret_value() == "legacy-rt"
        assert s.oauth2.access_token.get_secret_value() == "legacy-at"

        # other keys are not taken
        assert s.oauth2.scope == "offline_access"
        assert s.oauth2.audience == "https://api.earthscope.org"
        assert str(s.oauth2.domain) == "https://login.earthscope.org/"
        assert s.oauth2.client_id == "b9DtAFBd6QvMg761vI3YhYquNZbJX5G0"
        assert s.oauth2.scope == "offline_access"
        assert str(s.resources.api_url) == "https://api.earthscope.org/"
        assert s.http.user_agent == f"earthscope-sdk py/{__version__}"
        assert s.http.max_connections is None
        assert s.http.max_keepalive_connections is None
        assert s.http.keepalive_expiry.total_seconds() == 5.0
        assert s.http.timeout_read.total_seconds() == 5.0
        assert s.oauth2.client_secret is None

    def test_nested_settings_precedence(
        self, config_toml: Path, monkeypatch: MonkeyPatch
    ):
        config_toml.write_text(
            dedent("""
                [default]
                http.max_connections = 11
                http.max_keepalive_connections = 11
                http.timeout_read = 11.1
                http.keepalive_expiry = 11.1
                   
                http.retry.attempts = 11
                http.retry.timeout = 11.1
                http.retry.wait_jitter = 11.1
                http.retry.wait_initial = 11.1

                [profile.pytest]
                http.max_connections = 22
                http.max_keepalive_connections = 22
                http.keepalive_expiry = 22.2
                http.retry.timeout = 22.2
                http.retry.wait_jitter = 22.2
                http.retry.wait_initial = 22.2
            """)
        )

        monkeypatch.setenv("ES_HTTP__MAX_CONNECTIONS", "33")
        monkeypatch.setenv("ES_HTTP__KEEPALIVE_EXPIRY", "33.3")
        monkeypatch.setenv("ES_HTTP__RETRY__WAIT_JITTER", "33.3")
        monkeypatch.setenv("ES_HTTP__RETRY__WAIT_INITIAL", "33.3")

        s = SdkSettings(
            profile_name="pytest",
            http=HttpSettings(
                keepalive_expiry=44.4,
                retry=HttpRetrySettings(wait_initial=44.4),
            ),
        )

        assert s.http.keepalive_expiry.total_seconds() == 44.4
        assert s.http.max_connections == 33
        assert s.http.max_keepalive_connections == 22
        assert s.http.timeout_read.total_seconds() == 11.1

        assert s.http.retry.attempts == 11
        assert s.http.retry.timeout.total_seconds() == 22.2
        assert s.http.retry.wait_jitter.total_seconds() == 33.3
        assert s.http.retry.wait_initial.total_seconds() == 44.4


class TestTokens:
    def test_defaults(self):
        t = Tokens(access_token="at")
        assert t.refresh_token is None
        assert t.id_token is None

        t2 = Tokens(refresh_token="rt")
        assert t2.access_token is None

    def test_one_of_access_or_refresh(self):
        assert Tokens(access_token="at").access_token.get_secret_value() == "at"
        assert Tokens(refresh_token="rt").refresh_token.get_secret_value() == "rt"

        with pytest.raises(ValidationError):
            Tokens(id_token="it")

    def test_access_token_bad_body(self):
        with pytest.raises(ValueError):
            Tokens(access_token="at").access_token_body

    def test_access_token_body(self):
        # expired access token
        at = "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6Im80WDNMM1p0QkN6MmZ5RktMVW9mWiJ9.eyJpc3MiOiJodHRwczovL2xvZ2luLmVhcnRoc2NvcGUub3JnLyIsInN1YiI6Imdvb2dsZS1vYXV0aDJ8MTE4NDAzNTA2MTU0OTQyODQyNTgzIiwiYXVkIjoiaHR0cHM6Ly9hcGkuZWFydGhzY29wZS5vcmciLCJpYXQiOjE3MzQ1Nzg5MzQsImV4cCI6MTczNDY2NTMzNCwic2NvcGUiOiJvZmZsaW5lX2FjY2VzcyIsImF6cCI6ImI5RHRBRkJkNlF2TWc3NjF2STNZaFlxdU5aYkpYNUcwIn0.Sz9v95Iru62STcDfZYDXYiqW2zqVsAQ9_XpSaGgWbcXbJzQApG1pPQKL0s7yBw08diepEhUG4dOijdUKugw616kp7TkpH8hniYv7yPwwwdeqZYGTCJ0gfDIzxRYzPXoY1gqPM1WzA-ZPBNOdJyi_LNGS0_jjx4MGJ5NRkL76Us0zP-C9z_IdYA1jRgPrVkOt5AXSwFHNn-4FNDYwL2NE_eO2xfJZm4fHbhSmtWCT7wKJvt4g-viMprOiKK0glf9u2eDxXCP1QFIoNSug1C_esJAe1AbyaRlk0C3y8Zz42K8jMFurCIjI_PenQ4GAH0eunCSwrMPioIZq8k3a4MoKVw"
        t = Tokens(access_token=at)

        t.access_token_body.audience == "https://api.earthscope.org"
        t.access_token_body.client_id == "b9DtAFBd6QvMg761vI3YhYquNZbJX5G0"
        t.access_token_body.issuer == "https://login.earthscope.org/"
        t.access_token_body.subject == "google-oauth2|118403506154942842583"
        t.access_token_body.scope == "offline_access"
        t.access_token_body.grant_type is None

    def test_tokens_loading(self, default_settings: SdkSettings):
        exp = Tokens(access_token="at", id_token="it", refresh_token="rt")
        default_settings.write_tokens(exp)

        # load new settings as if
        actual = SdkSettings()
        assert (
            actual.oauth2.access_token.get_secret_value()
            == exp.access_token.get_secret_value()
        )
        assert (
            actual.oauth2.id_token.get_secret_value() == exp.id_token.get_secret_value()
        )
        assert (
            actual.oauth2.refresh_token.get_secret_value()
            == exp.refresh_token.get_secret_value()
        )


class TestTokensPrecedence:
    def test_tokens_precedence_all(
        self,
        config_toml: Path,
        default_settings: SdkSettings,
        monkeypatch: MonkeyPatch,
    ):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        default_settings.write_tokens(Tokens(access_token="state-at"))
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.access_token = "default-at"

                [profile.default]
                oauth2.access_token = "profile-at"
            """)
        )
        monkeypatch.setenv("ES_OAUTH2__ACCESS_TOKEN", "env-at")

        t = SdkSettings(oauth2=AuthFlowSettings(access_token="init-at"))
        assert t.oauth2.access_token.get_secret_value() == "init-at"

    def test_tokens_precedence_env(
        self,
        config_toml: Path,
        default_settings: SdkSettings,
        monkeypatch: MonkeyPatch,
    ):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        default_settings.write_tokens(Tokens(access_token="state-at"))
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.access_token = "default-at"

                [profile.default]
                oauth2.access_token = "profile-at"
            """)
        )
        monkeypatch.setenv("ES_OAUTH2__ACCESS_TOKEN", "env-at")

        t = SdkSettings()
        assert t.oauth2.access_token.get_secret_value() == "env-at"

    def test_tokens_precedence_profile(
        self,
        config_toml: Path,
        default_settings: SdkSettings,
        monkeypatch: MonkeyPatch,
    ):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        default_settings.write_tokens(Tokens(access_token="state-at"))
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.access_token = "default-at"

                [profile.default]
                oauth2.access_token = "profile-at"
            """)
        )

        t = SdkSettings()
        assert t.oauth2.access_token.get_secret_value() == "profile-at"

    def test_tokens_precedence_defaults(
        self,
        config_toml: Path,
        default_settings: SdkSettings,
        monkeypatch: MonkeyPatch,
    ):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        default_settings.write_tokens(Tokens(access_token="state-at"))
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.access_token = "default-at"
            """)
        )

        t = SdkSettings()
        assert t.oauth2.access_token.get_secret_value() == "default-at"

    def test_tokens_precedence_state(
        self,
        default_settings: SdkSettings,
        monkeypatch: MonkeyPatch,
    ):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)
        default_settings.write_tokens(Tokens(access_token="state-at"))

        t = SdkSettings()
        assert t.oauth2.access_token.get_secret_value() == "state-at"

    def test_tokens_precedence_legacy(
        self,
        config_toml: Path,
        legacy_tokens_json: Path,
        monkeypatch: MonkeyPatch,
    ):
        legacy_tokens_json.write_text(
            dedent("""
                {
                    "refresh_token": "legacy-rt",
                    "access_token": "legacy-at"
                }
            """)
        )

        # init > legacy
        s = SdkSettings(oauth2=AuthFlowSettings(access_token="init-at"))
        assert s.oauth2.refresh_token.get_secret_value() == "legacy-rt"
        assert s.oauth2.access_token.get_secret_value() == "init-at"

        # env > legacy
        with monkeypatch.context() as m:
            m.setenv("ES_OAUTH2__ACCESS_TOKEN", "env-at")

            s = SdkSettings()
            assert s.oauth2.refresh_token.get_secret_value() == "legacy-rt"
            assert s.oauth2.access_token.get_secret_value() == "env-at"

        # config.toml profile > default > legacy
        config_toml.write_text(
            dedent("""
                [default]
                oauth2.refresh_token = "default-rt"
                oauth2.access_token = "default-at"

                [profile.default]
                oauth2.access_token = "profile-at"
            """)
        )

        s = SdkSettings()
        assert s.oauth2.refresh_token.get_secret_value() == "default-rt"
        assert s.oauth2.access_token.get_secret_value() == "profile-at"
        config_toml.unlink()

        # legacy > bootstrap
        with monkeypatch.context() as m:
            m.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)

            s = SdkSettings()
            assert s.oauth2.refresh_token.get_secret_value() == "legacy-rt"
            assert s.oauth2.access_token.get_secret_value() == "legacy-at"
            assert s.oauth2.scope == "bootstrap-scope"


class TestBootstrapSettings:
    def test_bootstrap_settings(self, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, _bootstrap_state_json)

        s = SdkSettings()
        assert s.oauth2.audience == "https://bootstrap-audience.earthscope.org"
        assert s.oauth2.client_id == "bootstrap-client-id"
        assert str(s.oauth2.domain) == "https://bootstrap-domain.earthscope.org/"
        assert s.oauth2.scope == "bootstrap-scope"
        assert s.oauth2.access_token.get_secret_value() == "bootstrap-at"
        assert s.oauth2.refresh_token.get_secret_value() == "bootstrap-rt"
        assert s.oauth2.id_token.get_secret_value() == "bootstrap-it"

    def test_bootstrap_settings_invalid_json(self, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, "invalid-json")

        s = SdkSettings()
        assert s.oauth2.audience == "https://api.earthscope.org"
        assert s.oauth2.client_id == "b9DtAFBd6QvMg761vI3YhYquNZbJX5G0"
        assert str(s.oauth2.domain) == "https://login.earthscope.org/"
        assert s.oauth2.scope == "offline_access"
        assert s.oauth2.access_token is None
        assert s.oauth2.refresh_token is None
        assert s.oauth2.id_token is None

    def test_bootstrap_settings_empty_json(self, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(_BOOTSTRAP_ENV_VAR, "{}")

        s = SdkSettings()
        assert s.oauth2.audience == "https://api.earthscope.org"
        assert s.oauth2.client_id == "b9DtAFBd6QvMg761vI3YhYquNZbJX5G0"
        assert str(s.oauth2.domain) == "https://login.earthscope.org/"
        assert s.oauth2.scope == "offline_access"
        assert s.oauth2.access_token is None
        assert s.oauth2.refresh_token is None
        assert s.oauth2.id_token is None

    def test_bootstrap_settings_other_keys(self, monkeypatch: MonkeyPatch):
        monkeypatch.setenv(
            _BOOTSTRAP_ENV_VAR,
            json.dumps(
                {
                    "http": {
                        "timeout_read": 10.0,
                        "user_agent": "bootstrap-ua",
                    },
                    "oauth2": {
                        "scope": "bootstrap-scope",
                    },
                }
            ),
        )

        s = SdkSettings()
        # bootstrap settings
        assert s.http.timeout_read.total_seconds() == 10.0
        assert s.http.user_agent == "bootstrap-ua"
        assert s.oauth2.scope == "bootstrap-scope"

        # defaults
        assert s.http.timeout_connect.total_seconds() == 5.0
        assert s.oauth2.audience == "https://api.earthscope.org"
        assert s.oauth2.client_id == "b9DtAFBd6QvMg761vI3YhYquNZbJX5G0"
        assert str(s.oauth2.domain) == "https://login.earthscope.org/"
        assert s.oauth2.access_token is None
        assert s.oauth2.refresh_token is None
        assert s.oauth2.id_token is None


class TestAuthFlowSettings:
    @pytest.mark.parametrize(
        "host,allowed",
        [
            ("earthscope.org", True),
            ("api.earthscope.org", True),
            ("data.earthscope.org", True),
            ("foo.earthscope.org", True),
            ("foo.subdomain.earthscope.org", True),
            ("earthscope.foo.org", False),
            ("example.com", False),
            ("foo.example.com", False),
            ("earthscope.example.com", False),
            ("foo.earthscope.example.com", False),
        ],
    )
    def test_allowed_hosts_defaults(self, host: str, allowed: bool):
        s = AuthFlowSettings()
        assert s.is_host_allowed(host) == allowed

    def test_allowed_hosts_cache(self):
        s = AuthFlowSettings()
        assert "foo.earthscope.org" not in s.allowed_hosts
        assert s.is_host_allowed("foo.earthscope.org")
        assert "foo.earthscope.org" in s.allowed_hosts



import warnings

import httpx
import pytest
from pytest_httpx import HTTPXMock

from earthscope_sdk.auth.error import NoAccessTokenError
from earthscope_sdk.common.context import SdkContext
from earthscope_sdk.config.settings import SdkSettings


class TestContext:
    @pytest.mark.parametrize(
        "host",
        [
            "earthscope.org",
            "api.earthscope.org",
            "data.earthscope.org",
        ],
    )
    @pytest.mark.asyncio
    async def test_context_async_refreshes_and_injects_auth_header_for_allowed_hosts(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
        host: str,
    ):
        httpx_mock.add_response()

        ctx = SdkContext(mock_settings)
        with pytest.raises(NoAccessTokenError):
            ctx.auth_flow.access_token

        await ctx.httpx_client.get(f"https://{host}/foo")

        at = ctx.auth_flow.access_token
        req = httpx_mock.get_requests()

        assert req[1].url.host == host
        assert req[1].headers["authorization"] == f"Bearer {at}"
        assert req[1].headers["user-agent"] == mock_settings.http.user_agent

    @pytest.mark.parametrize(
        "host",
        [
            "foo.org",
            "earthscope.foo.org",
        ],
    )
    @pytest.mark.httpx_mock(assert_all_responses_were_requested=False)
    @pytest.mark.asyncio
    async def test_context_allowed_hosts_doesnt_inject_auth_header(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
        host: str,
    ):
        httpx_mock.add_response()

        ctx = SdkContext(mock_settings)
        with pytest.raises(NoAccessTokenError):
            ctx.auth_flow.access_token

        await ctx.httpx_client.get(f"https://{host}/bar")

        with pytest.raises(NoAccessTokenError):
            ctx.auth_flow.access_token

        req = httpx_mock.get_requests()

        assert req[0].url.host == host
        assert req[0].headers["user-agent"] == mock_settings.http.user_agent
        with pytest.raises(KeyError):
            req[0].headers["authorization"]

    @pytest.mark.asyncio
    async def test_context_async_close(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
    ):
        httpx_mock.add_response(text="OK")
        ctx = SdkContext(mock_settings)

        async def async_foo():
            return await ctx.httpx_client.get(f"{mock_settings.resources.api_url}foo")

        foo = ctx.syncify(async_foo)
        resp = foo()
        assert resp.text == "OK"

        assert not ctx.is_closed
        await ctx.async_close()

        assert ctx._httpx_client.is_closed
        assert ctx._runner.is_closed
        assert ctx.is_closed

        # can no longer schedule
        with warnings.catch_warnings(), pytest.raises(RuntimeError):
            warnings.filterwarnings(
                "ignore",
                message=r"coroutine '.+' was never awaited",
            )
            foo()

    def test_context_sync_close(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
    ):
        httpx_mock.add_response(text="OK")
        ctx = SdkContext(mock_settings)

        async def async_foo():
            return await ctx.httpx_client.get(f"{mock_settings.resources.api_url}foo")

        foo = ctx.syncify(async_foo)
        resp = foo()
        assert resp.text == "OK"

        assert not ctx.is_closed
        ctx.close()

        assert ctx._httpx_client.is_closed
        assert ctx._runner.is_closed
        assert ctx.is_closed

        # can no longer schedule
        with warnings.catch_warnings(), pytest.raises(RuntimeError):
            warnings.filterwarnings(
                "ignore",
                message=r"coroutine '.+' was never awaited",
            )
            foo()

    def test_auth_flow_in_custom_sync_client(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
    ):
        httpx_mock.add_response()

        ctx = SdkContext(mock_settings)
        with pytest.raises(NoAccessTokenError):
            ctx.auth_flow.access_token

        with httpx.Client(auth=ctx.auth_flow) as c:
            c.get(f"{mock_settings.resources.api_url}foo")

        at = ctx.auth_flow.access_token
        req = httpx_mock.get_requests()

        assert str(req[1].url).startswith(f"{mock_settings.resources.api_url}")
        assert req[1].headers["authorization"] == f"Bearer {at}"

        # note: in custom client, we don't expect our custom user-agent
        assert req[1].headers["user-agent"] != mock_settings.http.user_agent



import time
import webbrowser
from typing import Type

import pytest
from pytest_httpx import HTTPXMock

from earthscope_sdk.auth.client_credentials_flow import ClientCredentialsFlow
from earthscope_sdk.auth.device_code_flow import DeviceCodeFlow, PollingErrorType
from earthscope_sdk.auth.error import (
    DeviceCodePollingExpiredError,
    DeviceCodeRequestDeviceCodeError,
    NoRefreshTokenError,
    NoTokensError,
    UnauthorizedError,
)
from earthscope_sdk.client import AsyncEarthScopeClient
from earthscope_sdk.common.context import SdkContext
from earthscope_sdk.config.models import AuthFlowSettings, Tokens
from earthscope_sdk.config.settings import SdkSettings
from tests.conftest import AUDIENCE, CLIENT_ID, DOMAIN

from .util import (
    get_m2m_creds,
    get_refresh_token,
    is_pipeline,
    missing_m2m_creds,
    missing_refresh_token,
    register_mock_token_response,
    retries_enabled,
)


@pytest.fixture()
def device_flow_settings():
    return SdkSettings(
        oauth2=AuthFlowSettings(
            audience=AUDIENCE,
            domain=DOMAIN,
            client_id=CLIENT_ID,
        )
    )


@pytest.fixture()
def refresh_settings():
    return SdkSettings(
        oauth2=AuthFlowSettings(
            audience=AUDIENCE,
            domain=DOMAIN,
            client_id=CLIENT_ID,
            **get_refresh_token(),
        )
    )


@pytest.fixture()
def m2m_settings():
    return SdkSettings(
        oauth2=AuthFlowSettings(
            audience=AUDIENCE,
            domain=DOMAIN,
            **get_m2m_creds(),
        )
    )


class TestAuthDeviceCodeFlow:
    def test_no_tokens_device_code_flow(self, device_flow_settings: SdkSettings):
        flow = DeviceCodeFlow(SdkContext(device_flow_settings))

        with pytest.raises(NoTokensError):
            flow.access_token
        with pytest.raises(NoTokensError):
            flow.access_token_body
        with pytest.raises(NoTokensError):
            flow.refresh_token
        with pytest.raises(NoTokensError):
            flow.issued_at
        with pytest.raises(NoTokensError):
            flow.expires_at
        with pytest.raises(NoTokensError):
            flow.ttl
        with pytest.raises(NoTokensError):
            flow.scope

    @pytest.mark.asyncio
    async def test_device_code_request_error(self, httpx_mock: HTTPXMock):
        """Test handling of device code request errors."""
        httpx_mock.add_response(
            status_code=400,
            json={"error": "invalid_request", "error_description": "Invalid client"},
        )

        async with AsyncEarthScopeClient() as client:
            with pytest.raises(DeviceCodeRequestDeviceCodeError):
                await client.ctx.device_code_flow._async_request_device_code()

    @pytest.mark.parametrize(
        "err_code, ErrType",
        [
            (PollingErrorType.ACCESS_DENIED, UnauthorizedError),
            (PollingErrorType.EXPIRED_TOKEN, DeviceCodePollingExpiredError),
        ],
    )
    @pytest.mark.asyncio
    async def test_device_code_polling_error(
        self,
        httpx_mock: HTTPXMock,
        err_code: str,
        ErrType: Type[Exception],
    ):
        """Test handling of device code polling errors."""
        # First response for device code request
        httpx_mock.add_response(
            status_code=200,
            json={
                "device_code": "test_code",
                "user_code": "ABCD-EFGH",
                "verification_uri": "activate",
                "verification_uri_complete": "https://test.com/activate",
                "expires_in": 900,
                "interval": 0.001,
            },
        )
        # Second response for polling with error
        httpx_mock.add_response(
            status_code=400,
            json={
                "error": err_code,
                "error_description": "Pending",
            },
        )

        async with AsyncEarthScopeClient() as client:
            codes = await client.ctx.device_code_flow._async_request_device_code()

            with pytest.raises(ErrType):
                await client.ctx.device_code_flow._async_poll(codes=codes)

    @pytest.mark.skipif(
        is_pipeline(),
        reason="No user input in pipeline",
    )
    def test_login_device_code_flow(self, device_flow_settings: SdkSettings):
        flow = DeviceCodeFlow(SdkContext(device_flow_settings))

        with flow.do_flow() as code:
            webbrowser.open(code.verification_uri_complete)

        assert flow.access_token_body.audience == device_flow_settings.oauth2.audience
        assert flow.access_token_body.issuer == f"{device_flow_settings.oauth2.domain}"
        assert flow.access_token_body.scope == device_flow_settings.oauth2.scope
        assert flow.access_token_body.client_id == device_flow_settings.oauth2.client_id
        assert flow.refresh_token is not None

        raw = device_flow_settings.tokens_file.read_bytes()
        t = Tokens.model_validate_json(raw)
        assert t.access_token.get_secret_value() == flow.access_token
        assert t.refresh_token.get_secret_value() == flow.refresh_token

    @pytest.mark.skipif(
        missing_refresh_token(),
        reason="Missing refresh token",
    )
    def test_refresh_device_code_flow(self, refresh_settings: SdkSettings):
        flow = DeviceCodeFlow(SdkContext(refresh_settings))

        flow.refresh()

        at = flow.access_token
        body = flow.access_token_body
        assert flow.access_token_body.audience == refresh_settings.oauth2.audience
        assert flow.access_token_body.issuer == f"{refresh_settings.oauth2.domain}"
        assert flow.access_token_body.scope == refresh_settings.oauth2.scope
        assert flow.access_token_body.client_id == refresh_settings.oauth2.client_id
        assert (
            flow.refresh_token
            == refresh_settings.oauth2.refresh_token.get_secret_value()
        )

        time.sleep(1)

        flow.refresh()
        assert flow.access_token != at
        assert flow.access_token_body.issued_at > body.issued_at

    @pytest.mark.skipif(
        missing_refresh_token(),
        reason="Missing refresh token",
    )
    def test_refresh_if_necessary(self, refresh_settings: SdkSettings):
        flow = DeviceCodeFlow(SdkContext(refresh_settings))

        flow.refresh()

        at = flow.access_token
        time.sleep(1)

        flow.refresh_if_necessary()
        assert flow.access_token == at

    @pytest.mark.asyncio
    async def test_refresh_retry(
        self,
        httpx_mock: HTTPXMock,
    ):
        httpx_mock.add_response(429)
        httpx_mock.add_response(500)
        httpx_mock.add_response(502)
        httpx_mock.add_response(504)
        mock_token_body = register_mock_token_response(httpx_mock)

        settings = SdkSettings(
            oauth2=AuthFlowSettings(
                audience=AUDIENCE,
                domain=DOMAIN,
                refresh_token="mock-refresh-token",
            )
        )
        flow = DeviceCodeFlow(SdkContext(settings))

        with retries_enabled(5):
            await flow.async_refresh()

        assert len(httpx_mock.get_requests()) == 5
        assert flow.refresh_token == "mock-refresh-token"
        assert flow.access_token_body == mock_token_body


class TestAuthClientCredentialsFlow:
    def test_no_device_code_with_m2m_creds(self):
        settings = SdkSettings(
            oauth2=AuthFlowSettings(client_id="foo", client_secret="bar")
        )
        with pytest.raises(ValueError):
            DeviceCodeFlow(SdkContext(settings))

    def test_no_tokens_client_credentials_flow(self):
        settings = SdkSettings(
            oauth2=AuthFlowSettings(client_id="foo", client_secret="bar")
        )
        flow = ClientCredentialsFlow(SdkContext(settings))

        with pytest.raises(NoTokensError):
            flow.access_token
        with pytest.raises(NoTokensError):
            flow.access_token_body
        with pytest.raises(NoTokensError):
            flow.refresh_token
        with pytest.raises(NoTokensError):
            flow.issued_at
        with pytest.raises(NoTokensError):
            flow.expires_at
        with pytest.raises(NoTokensError):
            flow.ttl
        with pytest.raises(NoTokensError):
            flow.scope

    @pytest.mark.asyncio
    async def test_client_credentials_unauthorized(self, httpx_mock: HTTPXMock):
        settings = SdkSettings(
            oauth2=AuthFlowSettings(client_id="foo", client_secret="bar")
        )
        flow = ClientCredentialsFlow(SdkContext(settings))

        httpx_mock.add_response(status_code=401, json={})

        with pytest.raises(UnauthorizedError):
            await flow.async_request_tokens()

    @pytest.mark.skipif(
        missing_m2m_creds(),
        reason="Missing M2M credentials",
    )
    def test_login_client_credentials_flow(self, m2m_settings: SdkSettings):
        flow = ClientCredentialsFlow(SdkContext(m2m_settings))

        flow.refresh()

        assert flow.access_token_body.audience == m2m_settings.oauth2.audience
        assert flow.access_token_body.issuer == f"{m2m_settings.oauth2.domain}"
        # m2m scope depends on IdP setup
        assert flow.access_token_body.scope is not None
        assert flow.access_token_body.client_id == m2m_settings.oauth2.client_id

        with pytest.raises(NoRefreshTokenError):
            flow.refresh_token

        raw = m2m_settings.tokens_file.read_bytes()
        t = Tokens.model_validate_json(raw)
        assert t.access_token.get_secret_value() == flow.access_token
        assert t.refresh_token is None

    @pytest.mark.skipif(
        missing_m2m_creds(),
        reason="Missing M2M credentials",
    )
    def test_refresh_if_necessary(self, m2m_settings: SdkSettings):
        flow = ClientCredentialsFlow(SdkContext(m2m_settings))

        flow.refresh()

        at = flow.access_token
        time.sleep(1)

        flow.refresh_if_necessary()
        assert flow.access_token == at



import sys
from functools import cached_property
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Optional, TypeVar, cast

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec


if TYPE_CHECKING:
    from httpx import AsyncClient

    from earthscope_sdk.common._sync_runner import SyncRunner
    from earthscope_sdk.config.settings import SdkSettings

T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")


class SdkContext:
    """
    Shared context for the EarthScope SDK
    """

    @cached_property
    def auth_flow(self):
        """
        Reusable AuthFlow instance for managing token lifecycle
        """
        # lazy import; avoid circular dependency
        from earthscope_sdk.auth.auth_flow import AuthFlow
        from earthscope_sdk.config.models import AuthFlowType

        auth_flow_type = self.settings.oauth2.auth_flow_type

        if auth_flow_type == AuthFlowType.DeviceCode:
            return cast(AuthFlow, self.device_code_flow)

        if auth_flow_type == AuthFlowType.MachineToMachine:
            return cast(AuthFlow, self.client_credentials_flow)

        # Fall back to base auth flow; if a refresh token is present, we can still
        # facilitate automated renewal.
        return AuthFlow(ctx=self)

    @cached_property
    def client_credentials_flow(self):
        """
        Reusable ClientCredentialsFlow instance for managing token lifecycle
        """
        # lazy import; avoid circular dependency
        from earthscope_sdk.auth.client_credentials_flow import ClientCredentialsFlow

        return ClientCredentialsFlow(ctx=self)

    @cached_property
    def device_code_flow(self):
        """
        Resusable DeviceCodeFlow instance for managing token lifecycle
        """
        # lazy import; avoid circular dependency
        from earthscope_sdk.auth.device_code_flow import DeviceCodeFlow

        return DeviceCodeFlow(ctx=self)

    @cached_property
    def httpx_client(self):
        """
        Reusable HTTPx client for shared session and connection pooling.

        Automatically:
        - injects common headers (e.g. authorization, user-agent)
            - refreshes access token automatically
        """
        import httpx  # lazy import

        self._httpx_client = httpx.AsyncClient(
            auth=self.auth_flow,
            headers={
                "user-agent": self.settings.http.user_agent,
            },
            limits=self.settings.http.limits,
            timeout=self.settings.http.timeouts,
        )

        return self._httpx_client

    @property
    def is_closed(self):
        if (c := self._httpx_client) and not c.is_closed:
            return False

        if (r := self._runner) and not r.is_closed:
            return False

        return True

    @cached_property
    def settings(self):
        """
        Profile-specific settings merged with any configured defaults
        """
        return self._settings

    def __init__(
        self,
        settings: Optional["SdkSettings"] = None,
        *,
        runner: Optional["SyncRunner"] = None,
    ):
        # lazy import
        from earthscope_sdk.config.settings import SdkSettings

        # Local state
        self._httpx_client: Optional["AsyncClient"] = None
        self._runner: Optional["SyncRunner"] = runner
        self._settings = settings or SdkSettings()

    async def async_close(self):
        """
        Close this SdkContext to release underlying resources (e.g. connection pools)
        """
        if self._httpx_client:
            await self._httpx_client.aclose()

        if self._runner:
            self._runner.stop()

    def close(self):
        """
        Close this SdkContext to release underlying resources (e.g. connection pools)
        """
        # needs a different implementation than async_close()

        # need to run aclose() in the event loop
        if self._httpx_client:
            self.syncify(self._httpx_client.aclose)()

        # closing the event loop should be done *outside* the event loop *after* all async cleanup
        if self._runner:
            self._runner.stop()

    def syncify(
        self,
        async_function: Callable[T_ParamSpec, Coroutine[Any, Any, T_Retval]],
    ) -> Callable[T_ParamSpec, T_Retval]:
        """
        Create a synchronous version of an async function
        """
        if not self._runner:
            self._runner = self._get_runner()

        return self._runner.syncify(async_function)

    def _get_runner(self) -> "SyncRunner":
        # lazy imports
        import asyncio

        from earthscope_sdk.common._sync_runner import (
            BackgroundSyncRunner,
            SimpleSyncRunner,
        )

        try:
            asyncio.get_running_loop()

        except RuntimeError:
            # No event loop exists in this thread; we can run in this thread
            return SimpleSyncRunner()

        else:
            # An event loop already exists in this thread; create a background thread
            return BackgroundSyncRunner()



import datetime as dt
from typing import Optional

from pydantic import BaseModel, ConfigDict


class UserProfile(BaseModel):
    """
    EarthScope user profile
    """

    first_name: str
    last_name: str
    country_code: str
    region_code: Optional[str]
    institution: str
    work_sector: str
    user_id: str
    primary_email: str
    created_at: dt.datetime
    updated_at: dt.datetime


class AwsTemporaryCredentials(BaseModel):
    """
    AWS temporary credentials
    """

    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str
    expiration: dt.datetime

    model_config = ConfigDict(frozen=True)

    @property
    def ttl(self):
        """
        Time to live
        """
        return self.expiration - dt.datetime.now(dt.timezone.utc)

    def is_expired(self, ttl_threshold: Optional[dt.timedelta] = None):
        """
        Check if these credentials are past or near expiration

        Args:
            ttl_threshold: remaining time-to-live before considering the temporary AWS creds expired (Default: 30s)
        """
        if ttl_threshold is None:
            ttl_threshold = dt.timedelta(seconds=30)

        return self.ttl <= ttl_threshold



import abc
import asyncio
import functools
import sys
from typing import Any, Callable, Coroutine, TypeVar

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")


class SyncRunner(abc.ABC):
    """
    Facilitates running async functions in a synchronous manner
    """

    @property
    def is_closed(self) -> bool:
        """
        The async runner is closed and no more synchronous invocations may occur.
        """
        return False

    @abc.abstractmethod
    def stop(self):
        """
        Stop the async runner (i.e. clean up any allocated resources).

        Methods created via `.syncify()` may not work after stopping this runner.
        """

    @abc.abstractmethod
    def syncify(
        self,
        async_function: Callable[T_ParamSpec, Coroutine[Any, Any, T_Retval]],
    ) -> Callable[T_ParamSpec, T_Retval]:
        """
        Create a synchronous version of an async function
        """


class SimpleSyncRunner(SyncRunner):
    """
    Runs async functions in a dedicated event loop in the same thread as the caller.

    This is a simple light-weight option, but does not work when an EventLoop already
    exists in the calling thread.
    """

    def __init__(self):
        self._loop = asyncio.new_event_loop()

    @property
    def is_closed(self):
        return self._loop.is_closed()

    def stop(self):
        # implements abstract method
        self._loop.stop()
        self._loop.close()

    def syncify(
        self,
        async_function: Callable[T_ParamSpec, Coroutine[Any, Any, T_Retval]],
    ) -> Callable[T_ParamSpec, T_Retval]:
        # implements abstract method

        @functools.wraps(async_function)
        def wrapper(*args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs) -> T_Retval:
            partial_f = functools.partial(async_function, *args, **kwargs)

            # run_until_complete() only runs the event loop until partial_f() has finished.
            # - The loop does not continue running in the background.
            # - This method will throw a RuntimeError if there is already an event loop
            #   running in the calling thread.
            return self._loop.run_until_complete(partial_f())

        return wrapper


class BackgroundSyncRunner(SyncRunner):
    """
    Runs async functions in a dedicated event loop in a background thread.

    This has the overhead of spawning a dedicated daemon thread.
    """

    def __init__(self):
        # lazy import
        import threading

        self._lock = threading.Lock()
        self._loop = asyncio.new_event_loop()
        self._running = False
        self._thread = threading.Thread(target=self._run_loop, daemon=True)

    @property
    def is_closed(self):
        return self._loop.is_closed() and not self._thread.is_alive()

    def start(self):
        with self._lock:
            if not self._thread.is_alive():
                self._thread.start()

    def stop(self):
        # implements abstract method
        with self._lock:
            if self._running:
                self._loop.call_soon_threadsafe(self._loop.stop)
                self._thread.join()

    def syncify(
        self,
        async_function: Callable[T_ParamSpec, Coroutine[Any, Any, T_Retval]],
    ) -> Callable[T_ParamSpec, T_Retval]:
        # implements abstract method

        # start the thread (if necessary) for running sync functions created here
        self.start()

        @functools.wraps(async_function)
        def wrapper(*args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs) -> T_Retval:
            partial_f = functools.partial(async_function, *args, **kwargs)
            fut = asyncio.run_coroutine_threadsafe(partial_f(), self._loop)
            return fut.result()

        return wrapper

    def _run_loop(self):
        asyncio.set_event_loop(self._loop)
        self._running = True
        try:
            self._loop.run_forever()
        finally:
            self._loop.close()
            self._running = False



"""
This module facilitates compatibility with EarthScope CLI v0.x.x
"""

import json
import logging
import os
import sys
from pathlib import Path
from typing import Union

from pydantic_settings import PydanticBaseSettingsSource

logger = logging.getLogger(__name__)

####
# The following code is nearly verbatim from Click's source (which is what Typer uses underneath)
# https://github.com/pallets/click/blob/main/src/click/utils.py.
#
# The various functions and consts have been renamed to be "private"
#
# Click's get_app_dir() was used to determine the "application directory" in earthscope-cli v0.x.x.
# whereas this has moved to ~/.earthscope/ starting in 1.0.0
#
# We replicate this functionality here so that the SDK does not need a dependency on Click.
####

_WIN = sys.platform.startswith("win")


def _posixify(name: str) -> str:
    return "-".join(name.split()).lower()


def _get_legacy_app_dir(
    app_name: str,
    roaming: bool = True,
    force_posix: bool = False,
) -> str:
    r"""Returns the config folder for the application.  The default behavior
    is to return whatever is most appropriate for the operating system.

    To give you an idea, for an app called ``"Foo Bar"``, something like
    the following folders could be returned:

    Mac OS X:
      ``~/Library/Application Support/Foo Bar``
    Mac OS X (POSIX):
      ``~/.foo-bar``
    Unix:
      ``~/.config/foo-bar``
    Unix (POSIX):
      ``~/.foo-bar``
    Windows (roaming):
      ``C:\Users\<user>\AppData\Roaming\Foo Bar``
    Windows (not roaming):
      ``C:\Users\<user>\AppData\Local\Foo Bar``

    .. versionadded:: 2.0

    :param app_name: the application name.  This should be properly capitalized
                     and can contain whitespace.
    :param roaming: controls if the folder should be roaming or not on Windows.
                    Has no effect otherwise.
    :param force_posix: if this is set to `True` then on any POSIX system the
                        folder will be stored in the home folder with a leading
                        dot instead of the XDG config home or darwin's
                        application support folder.
    """
    if _WIN:
        key = "APPDATA" if roaming else "LOCALAPPDATA"
        folder = os.environ.get(key)
        if folder is None:
            folder = os.path.expanduser("~")
        return os.path.join(folder, app_name)
    if force_posix:
        return os.path.join(os.path.expanduser(f"~/.{_posixify(app_name)}"))
    if sys.platform == "darwin":
        return os.path.join(
            os.path.expanduser("~/Library/Application Support"), app_name
        )
    return os.path.join(
        os.environ.get("XDG_CONFIG_HOME", os.path.expanduser("~/.config")),
        _posixify(app_name),
    )


####
# end of Click code
####

# Legacy state was just the exact response body from Auth0 e.g.
# {
#   "access_token": "<TOKEN>",
#   "id_token": "<TOKEN>",
#   "refresh_token": "<TOKEN>",
#   "scope": "openid profile email offline_access",
#   "expires_at": 1734742981,
#   "issued_at": 1734714181
# }

# the following is how the local app directory was determined in v0.x.x
_LEGACY_APP_DIR = Path(
    os.environ.get("APP_DIRECTORY", _get_legacy_app_dir("earthscope-cli"))
)


def _get_legacy_auth_state_path():
    return _LEGACY_APP_DIR / "sso_tokens.json"


class LegacyEarthScopeCLISettingsSource(PydanticBaseSettingsSource):
    """
    This SettingsSource facilitates migrating earthscope-cli v0.x.x local state into
    earthscope-sdk v1.x.x state
    """

    def __init__(self, settings_cls, *keys: str):
        super().__init__(settings_cls)
        self._keys = keys

    def __call__(self):
        # load path dynamically so we can override in tests
        state_path = _get_legacy_auth_state_path()

        # attempt to load from legacy sso_tokens.json
        try:
            with state_path.open() as f:
                state: dict[str, Union[str, int]] = json.load(f)

        # In either error case, just return an empty
        except json.JSONDecodeError:
            logger.warning(
                f"Found legacy earthscope-cli state at '{state_path}', but unable to decode content as JSON"
            )
            return {}

        except FileNotFoundError:
            return {}

        # only preserve specific keys if non-null
        settings = {k: v for k in self._keys if (v := state.get(k)) is not None}
        return {"oauth2": settings}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def get_field_value(self, *args, **kwargs): ...  # unused abstract method



import datetime as dt

import pytest
from pytest_httpx import HTTPXMock

from earthscope_sdk.auth.error import NoAccessTokenError
from earthscope_sdk.client import AsyncEarthScopeClient, EarthScopeClient
from earthscope_sdk.client.user.models import AwsTemporaryCredentials, UserProfile
from earthscope_sdk.common.context import SdkContext
from earthscope_sdk.config.settings import SdkSettings
from tests.util import retries_enabled


class TestSyncClient:
    def test_sdk_auto_injects_auth_header(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
    ):
        u = UserProfile(
            first_name="Jane",
            last_name="Doe",
            country_code="US",
            region_code="CO",
            institution="EarthScope Consortium",
            work_sector="non-profit",
            user_id="user-id-123",
            primary_email="jane.doe@earthscope.org",
            created_at="2024-01-01T00:00:00Z",
            updated_at="2024-03-01T00:00:00Z",
        )
        httpx_mock.add_response(json=u.model_dump(mode="json"))

        with EarthScopeClient(settings=mock_settings) as client:
            with pytest.raises(NoAccessTokenError):
                client.ctx.auth_flow.access_token

            u_resp = client.user.get_profile()

            at = client.ctx.auth_flow.access_token

        req = httpx_mock.get_requests()

        assert str(req[1].url).startswith(f"{mock_settings.resources.api_url}")
        assert req[1].headers["authorization"] == f"Bearer {at}"
        assert req[1].headers["user-agent"] == mock_settings.http.user_agent

        assert u == u_resp

        assert client.is_closed


class TestAsyncClient:
    @pytest.mark.asyncio
    async def test_sdk_auto_injects_auth_header(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
    ):
        u = UserProfile(
            first_name="Jane",
            last_name="Doe",
            country_code="US",
            region_code="CO",
            institution="EarthScope Consortium",
            work_sector="non-profit",
            user_id="user-id-123",
            primary_email="jane.doe@earthscope.org",
            created_at="2024-01-01T00:00:00Z",
            updated_at="2024-03-01T00:00:00Z",
        )
        httpx_mock.add_response(json=u.model_dump(mode="json"))

        async with AsyncEarthScopeClient(settings=mock_settings) as client:
            with pytest.raises(NoAccessTokenError):
                client.ctx.auth_flow.access_token

            u_resp = await client.user.get_profile()

            at = client.ctx.auth_flow.access_token

        req = httpx_mock.get_requests()

        assert str(req[1].url).startswith(f"{mock_settings.resources.api_url}")
        assert req[1].headers["authorization"] == f"Bearer {at}"
        assert req[1].headers["user-agent"] == mock_settings.http.user_agent

        assert u == u_resp

        assert client.is_closed

    @pytest.mark.asyncio
    async def test_aws_credentials_caching(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
    ):
        # reuse context
        ctx = SdkContext(settings=mock_settings)

        now = dt.datetime.now(dt.timezone.utc)
        c1 = AwsTemporaryCredentials(
            aws_access_key_id="mock-access-key",
            aws_secret_access_key="mock-secret-key",
            aws_session_token="mock-session-token",
            expiration=now + dt.timedelta(hours=1),  # creds must expire in the future
        )
        httpx_mock.add_response(json=c1.model_dump(mode="json"))

        c2 = AwsTemporaryCredentials(
            aws_access_key_id="mock-access-key-refresh",
            aws_secret_access_key="mock-secret-key-refresh",
            aws_session_token="mock-session-token-refresh",
            expiration=now + dt.timedelta(hours=2),  # creds must expire in the future
        )
        httpx_mock.add_response(json=c2.model_dump(mode="json"))

        async with AsyncEarthScopeClient(ctx=ctx) as client:
            c_resp = await client.user.get_aws_credentials()
            assert c_resp == c1, "got credentials"

            c_resp = await client.user.get_aws_credentials()
            assert c_resp == c1, "memory cache hit"

        # create a new client with empty memory cache
        async with AsyncEarthScopeClient(ctx=ctx) as client:
            c_resp = await client.user.get_aws_credentials()
            assert c_resp == c1, "disk cache hit"

            c_resp = await client.user.get_aws_credentials(force=True)
            assert c_resp == c2, "got new credentials when forced"

    @pytest.mark.asyncio
    async def test_retries(
        self,
        mock_settings: SdkSettings,
        httpx_mock: HTTPXMock,
    ):
        # reuse context
        ctx = SdkContext(settings=mock_settings)

        httpx_mock.add_response(429)
        httpx_mock.add_response(429)

        u = UserProfile(
            first_name="Jane",
            last_name="Doe",
            country_code="US",
            region_code="CO",
            institution="EarthScope Consortium",
            work_sector="non-profit",
            user_id="user-id-123",
            primary_email="jane.doe@earthscope.org",
            created_at="2024-01-01T00:00:00Z",
            updated_at="2024-03-01T00:00:00Z",
        )
        httpx_mock.add_response(json=u.model_dump(mode="json"))

        with retries_enabled(3):
            async with AsyncEarthScopeClient(ctx=ctx) as client:
                u_resp = await client.user.get_profile()
                assert u == u_resp



class AuthFlowError(Exception):
    """Generic authentication flow error"""


class UnauthenticatedError(AuthFlowError):
    pass


class UnauthorizedError(AuthFlowError):
    pass


class NoTokensError(AuthFlowError):
    pass


class NoAccessTokenError(NoTokensError):
    pass


class NoIdTokenError(NoTokensError):
    pass


class NoRefreshTokenError(NoTokensError):
    pass


class InvalidRefreshTokenError(AuthFlowError):
    pass


class ClientCredentialsFlowError(AuthFlowError):
    pass


class DeviceCodeRequestDeviceCodeError(AuthFlowError):
    pass


class DeviceCodePollingError(AuthFlowError):
    pass


class DeviceCodePollingExpiredError(DeviceCodePollingError):
    pass



import pathlib
from functools import reduce


def _merge(a: dict, b: dict):
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                _merge(a[key], b[key])
                continue

        a[key] = b[key]

    return a


def deep_merge(*mappings: dict, **kwargs):
    """
    Merge all of the dictionaries provided, reursing into nested dictionaries.

    Behavior:
    - The precedence increases from left to right, with keyword arguments as the highest precedence.
    - The leftmost dictionary is modified in-place. Pass an empty dictionary first to avoid modification of inputs.
    """
    return reduce(_merge, (*mappings, kwargs))


def get_config_dir(app_name: str):
    r"""
    Returns the hidden config folder in the user's home directory for the application.

    Mac OS X:
        `~/.app-name`
    Unix:
        `~/.app-name`
    Windows:
        `C:\Users\<user>\.app-name`
    """
    app_slug = slugify(app_name)
    config_dir = pathlib.Path.home() / f".{app_slug}"
    return config_dir


def slugify(name: str) -> str:
    """
    Convert the given name into a slug which can be used more easily in file names
    """
    return "-".join(name.split()).lower()



from typing import TYPE_CHECKING

from earthscope_sdk.auth.error import UnauthenticatedError, UnauthorizedError

if TYPE_CHECKING:
    from httpx import Request

    from earthscope_sdk.common.context import SdkContext


class SdkService:
    """
    Base class for EarthScope SDK services
    """

    @property
    def ctx(self):
        """
        SDK Context
        """
        return self._ctx

    @property
    def resources(self):
        """
        References to EarthScope resources
        """
        return self.ctx.settings.resources

    def __init__(self, ctx: "SdkContext"):
        self._ctx = ctx

    async def _send(self, request: "Request"):
        """
        Send an HTTP request.

        Performs common response handling.
        """
        resp = await self.ctx.httpx_client.send(request=request)

        # Throw specific errors for certain status codes

        if resp.status_code == 401:
            await resp.aread()  # must read body before using .text prop
            raise UnauthenticatedError(resp.text)

        if resp.status_code == 403:
            await resp.aread()  # must read body before using .text prop
            raise UnauthorizedError(resp.text)

        # Raise HTTP errors
        resp.raise_for_status()

        return resp

    async def _send_with_retries(self, request: "Request"):
        async for attempt in self.ctx.settings.http.retry.retry_context():
            with attempt:
                return await self._send(request=request)



import datetime as dt
import logging
from contextlib import suppress
from typing import Optional

import httpx
from pydantic import ValidationError

from earthscope_sdk.auth.error import (
    InvalidRefreshTokenError,
    NoAccessTokenError,
    NoRefreshTokenError,
    NoTokensError,
)
from earthscope_sdk.common.context import SdkContext
from earthscope_sdk.config.models import AccessTokenBody, Tokens

logger = logging.getLogger(__name__)


class AuthFlow(httpx.Auth):
    """
    Generic oauth2 flow class for handling retrieving, validating, saving, and refreshing access tokens.
    """

    @property
    def access_token(self):
        """
        Access token

        Raises:
            NoTokensError: no tokens (at all) are present
            NoAccessTokenError: no access token is present
        """
        if at := self.tokens.access_token:
            return at.get_secret_value()

        raise NoAccessTokenError("No access token was found. Please re-authenticate.")

    @property
    def access_token_body(self):
        """
        Access token body

        Raises:
            NoAccessTokenError: no access token is present
        """
        if body := self._access_token_body:
            return body

        raise NoAccessTokenError("No access token was found. Please re-authenticate.")

    @property
    def expires_at(self):
        """
        Time of access token expiration

        Raises:
            NoAccessTokenError: no access token is present
        """
        return self.access_token_body.expires_at

    @property
    def has_refresh_token(self):
        """
        Whether or not we have a refresh token
        """
        try:
            return self.refresh_token is not None
        except NoTokensError:
            return False

    @property
    def issued_at(self):
        """
        Time of access token creation

        Raises:
            NoAccessTokenError: no access token is present
        """
        return self.access_token_body.issued_at

    @property
    def refresh_token(self):
        """
        Refresh token

        Raises:
            NoTokensError: no tokens (at all) are present
            NoRefreshTokenError: no refresh token is present
        """
        if rt := self.tokens.refresh_token:
            return rt.get_secret_value()

        raise NoRefreshTokenError("No refresh token was found. Please re-authenticate.")

    @property
    def scope(self):
        """
        Access token scope

        Raises:
            NoAccessTokenError: no access token is present
        """
        return set(self.access_token_body.scope.split())

    @property
    def tokens(self):
        """
        Oauth2 tokens managed by this auth flow

        Raises:
            NoTokensError: no tokens (at all) are present
        """
        if tokens := self._tokens:
            return tokens

        raise NoTokensError("No tokens were found. Please re-authenticate.")

    @property
    def ttl(self):
        """
        Access token time-to-live (ttl) before expiration

        Raises:
            NoAccessTokenError: no access token is present
        """
        return self.access_token_body.ttl

    def __init__(self, ctx: SdkContext) -> None:
        self._ctx = ctx
        self._settings = ctx.settings.oauth2

        # Local state
        self._access_token_body: Optional[AccessTokenBody] = None
        self._tokens: Optional[Tokens] = None

        # httpx.Auth objects can be used in either sync or async clients so we
        # facilitate both from the same class
        self.refresh = ctx.syncify(self.async_refresh)
        self.refresh_if_necessary = ctx.syncify(self.async_refresh_if_necessary)
        self.revoke_refresh_token = ctx.syncify(self.async_revoke_refresh_token)

        # Initialize with provided tokens
        with suppress(ValidationError):
            self._validate_tokens(self._settings)

    async def async_refresh(self, scope: Optional[str] = None):
        """
        Refresh the access token

        Args:
            scope: the specific oauth2 scopes to request

        Raises:
            NoTokensError: no tokens (at all) are present
            NoRefreshTokenError: no refresh token is present
            InvalidRefreshTokenError: the token refresh failed
        """
        from httpx import HTTPStatusError, ReadTimeout

        refresh_token = self.refresh_token
        scope = scope or self._settings.scope

        request = self._ctx.httpx_client.build_request(
            "POST",
            f"{self._settings.domain}oauth/token",
            headers={"content-type": "application/x-www-form-urlencoded"},
            data={
                "grant_type": "refresh_token",
                "client_id": self._settings.client_id,
                "refresh_token": refresh_token,
                "scopes": scope,
            },
        )

        try:
            async for attempt in self._settings.retry.retry_context(ReadTimeout):
                with attempt:
                    r = await self._ctx.httpx_client.send(request, auth=None)
                    r.raise_for_status()
        except HTTPStatusError as e:
            logger.error(
                f"error during token refresh ({attempt.num} attempts): {e.response.content}"
            )
            raise InvalidRefreshTokenError("refresh token exchange failed")
        except Exception as e:
            logger.error(
                f"error during token refresh ({attempt.num} attempts)", exc_info=e
            )
            raise InvalidRefreshTokenError("refresh token exchange failed") from e

        # add previous refresh token to new tokens if omitted from resp
        # (i.e. we have a non-rotating refresh token)
        resp: dict = r.json()
        resp.setdefault("refresh_token", refresh_token)

        self._validate_and_save_tokens(resp)

        logger.debug(f"Refreshed tokens: {self._tokens}")
        return self

    async def async_refresh_if_necessary(
        self,
        scope: Optional[str] = None,
        auto_refresh_threshold: int = 60,
    ):
        """
        Refresh the access token if it is expired or bootstrap the access token from a refresh token

        Args:
            scope: the specific oauth2 scopes to request
            auto_refresh_threshold: access token TTL remaining (in seconds) before auto-refreshing

        Raises:
            NoTokensError: no tokens (at all) are present
            NoRefreshTokenError: no refresh token is present
            InvalidRefreshTokenError: the token refresh failed
        """
        # Suppress to allow bootstrapping with a refresh token
        with suppress(NoAccessTokenError):
            if self.ttl >= dt.timedelta(seconds=auto_refresh_threshold):
                return self

        return await self.async_refresh(scope=scope)

    async def async_revoke_refresh_token(self):
        """
        Revoke the refresh token.

        This invalidates the refresh token server-side so it may never again be used to
        get new access tokens.

        Raises:
            NoTokensError: no tokens (at all) are present
            NoRefreshTokenError: no refresh token is present
            InvalidRefreshTokenError: the token revocation failed
        """
        refresh_token = self.refresh_token

        r = await self._ctx.httpx_client.post(
            f"{self._settings.domain}oauth/revoke",
            auth=None,  # override client default
            headers={"content-type": "application/json"},
            json={
                "client_id": self._settings.client_id,
                "token": refresh_token,
            },
        )

        if r.status_code != 200:
            logger.error(f"error while revoking refresh token: {r.content}")
            raise InvalidRefreshTokenError("refresh token revocation failed")

        logger.debug(f"Refresh token revoked: {refresh_token}")
        return self

    def _validate_and_save_tokens(self, unvalidated_tokens: dict):
        """
        Validate then save tokens to local storage

        Args:
            unvalidated_tokens: tokens that have yet to be validated

        Returns:
            this auth flow

        Raises:
            pydantic.ValidationError: the token's body could not be decoded
        """
        self._validate_tokens(unvalidated_tokens)

        try:
            self._ctx.settings.write_tokens(self._tokens)
        except Exception as e:
            logger.error("Error while persisting tokens", exc_info=e)
            raise

        return self

    def _validate_tokens(self, unvalidated_tokens: dict):
        """
        Validate the tokens

        Args:
            unvalidated_tokens: tokens that have yet to be validated

        Returns:
            this auth flow

        Raises:
            pydantic.ValidationError: the token's body could not be decoded
        """

        tokens = Tokens.model_validate(unvalidated_tokens)

        self._access_token_body = tokens.access_token_body
        self._tokens = tokens

        return self

    ##########
    # The following methods make this class compatible with httpx.Auth.
    ##########

    async def async_auth_flow(self, request: httpx.Request):
        """
        Injects authorization into the request

        (this method makes this class httpx.Auth compatible)
        """
        super().async_auth_flow
        if request.headers.get("authorization") is None:
            if self._settings.is_host_allowed(request.url.host):
                await self.async_refresh_if_necessary()
                access_token = self.access_token
                request.headers["authorization"] = f"Bearer {access_token}"

        yield request

    def sync_auth_flow(self, request: httpx.Request):
        """
        Injects authorization into the request

        (this method makes this class httpx.Auth compatible)
        """
        # NOTE: we explicitly redefine this sync method because ctx.syncify()
        # does not support generators
        if request.headers.get("authorization") is None:
            if self._settings.is_host_allowed(request.url.host):
                self.refresh_if_necessary()
                access_token = self.access_token
                request.headers["authorization"] = f"Bearer {access_token}"

        yield request



import logging
from json import JSONDecodeError

from earthscope_sdk.auth.auth_flow import AuthFlow
from earthscope_sdk.auth.error import ClientCredentialsFlowError, UnauthorizedError
from earthscope_sdk.common.context import SdkContext

logger = logging.getLogger(__name__)


class ClientCredentialsFlow(AuthFlow):
    """
    Implements the oauth2 Client Credentials "machine-to-machine" (M2M) flow.
    """

    def __init__(self, ctx: SdkContext):
        if not ctx.settings.oauth2.client_secret:
            raise ValueError("Client secret required for client credentials flow")

        super().__init__(ctx=ctx)

        # httpx.Auth objects can be used in either sync or async clients so we
        # facilitate both from the same class
        self.request_tokens = ctx.syncify(self.async_request_tokens)

    async def async_refresh(self, *args, **kwargs):
        # alias for self.request_tokens() so that base class can get new tokens automagically
        return await self.async_request_tokens()

    async def async_request_tokens(self):
        """
        Request access token from IdP

        Raises:
            Unauthorized Error: the request returns as unauthorized
            ClientCredentialsFlowError: unhandled error in the client credentials flow
        """
        r = await self._ctx.httpx_client.post(
            f"{self._settings.domain}oauth/token",
            auth=None,  # override client default
            headers={"content-type": "application/x-www-form-urlencoded"},
            data={
                "grant_type": "client_credentials",
                "client_id": self._settings.client_id,
                "client_secret": self._settings.client_secret.get_secret_value(),
                "audience": self._settings.audience,
            },
        )
        try:
            resp: dict = r.json()
        except JSONDecodeError:
            raise ClientCredentialsFlowError(
                f"Unable to unpack IdP response: {r.content}"
            )

        # Success!
        if r.status_code == 200:
            self._validate_and_save_tokens(resp)

            logger.debug(f"Got tokens: {self.tokens}")
            return self

        # Unauthorized
        if r.status_code == 401:
            raise UnauthorizedError(
                f"m2m client '{self._settings.client_id}' is not authorized. IdP response: {resp}"
            )

        # Unhandled
        raise ClientCredentialsFlowError("client credentials flow failed", r.content)



import logging
from asyncio import sleep
from contextlib import asynccontextmanager, contextmanager
from enum import Enum
from json import JSONDecodeError
from typing import Optional

from pydantic import BaseModel, ValidationError

from earthscope_sdk.auth.auth_flow import AuthFlow
from earthscope_sdk.auth.error import (
    DeviceCodePollingError,
    DeviceCodePollingExpiredError,
    DeviceCodeRequestDeviceCodeError,
    UnauthorizedError,
)
from earthscope_sdk.common.context import SdkContext

logger = logging.getLogger(__name__)


class PollingErrorType(str, Enum):
    AUTHORIZATION_PENDING = "authorization_pending"
    SLOW_DOWN = "slow_down"
    EXPIRED_TOKEN = "expired_token"
    ACCESS_DENIED = "access_denied"


class GetDeviceCodeResponse(BaseModel):
    device_code: str
    user_code: str
    verification_uri: str
    verification_uri_complete: str
    expires_in: int
    interval: float


class PollingErrorResponse(BaseModel):
    error: PollingErrorType
    error_description: str


class DeviceCodeFlow(AuthFlow):
    """
    Implements the oauth2 Device Code flow.
    """

    @property
    def polling(self):
        """
        This instance is in the process of polling for tokens
        """
        return self._is_polling

    def __init__(self, ctx: SdkContext):
        if ctx.settings.oauth2.client_secret:
            raise ValueError("Client secret should not be used with device code flow")

        super().__init__(ctx=ctx)

        # Flow management vars
        self._is_polling = False

        # httpx.Auth objects can be used in either sync or async clients so we
        # facilitate both from the same class
        self._poll = ctx.syncify(self._async_poll)
        self._request_device_code = ctx.syncify(self._async_request_device_code)

    @asynccontextmanager
    async def async_do_flow(self, scope: Optional[str] = None):
        """
        Perform the oauth2 Device Code flow:
        - requests device code
        - yields the device code to the caller for them to prompt the user
        - polls for access token

        Args:
            scope: the specific oauth2 scopes to request

        Yields:
            GetDeviceCodeResponse: the device code response that must be shown to the user
                to facilitate a login.

        Raises:
            DeviceCodeRequestDeviceCodeError: failed to get a device code from the IdP
            DeviceCodePollingExpiredError: the polling token expired
            UnauthorizedError: access denied
            DeviceCodePollingError: unhandled polling error

        Examples:
            >>> async with device_flow.async_do_flow() as codes:
            ...     # prompt the user to login using the device code
            ...     print(f"Open the following URL in a web browser to login: {codes.verification_uri_complete}")
        """
        # NOTE: if you update this method, also update the synchronous version

        codes = await self._async_request_device_code(scope=scope)

        # Yield codes to facilitate prompting the user
        yield codes

        await self._async_poll(codes=codes)

    @contextmanager
    def do_flow(self, scope: Optional[str] = None):
        """
        Perform the oauth2 Device Code flow:
        - requests device code
        - yields the device code to the caller for them to prompt the user
        - polls for access token

        Args:
            scope: the specific oauth2 scopes to request

        Yields:
            GetDeviceCodeResponse: the device code response that must be shown to the user
                to facilitate a login.

        Raises:
            DeviceCodeRequestDeviceCodeError: failed to get a device code from the IdP
            DeviceCodePollingExpiredError: the polling token expired
            UnauthorizedError: access denied
            DeviceCodePollingError: unhandled polling error

        Examples:
            >>> with device_flow.do_flow() as codes:
            ...     # prompt the user to login using the device code
            ...     print(f"Open the following URL in a web browser to login: {codes.verification_uri_complete}")
        """
        # NOTE: we explicitly redefine this sync method because ctx.syncify()
        # does not support generators

        codes = self._request_device_code(scope=scope)

        # Yield codes to facilitate prompting the user
        yield codes

        self._poll(codes=codes)

    async def _async_poll(self, codes: GetDeviceCodeResponse):
        """
        Polling IdP for tokens.

        Args:
            codes: device code response from IdP after starting the flow

        Raises:
            DeviceCodePollingExpiredError: the polling token expired
            UnauthorizedError: access denied
            DeviceCodePollingError: unhandled polling error
        """
        if self._is_polling:
            raise DeviceCodePollingError("Polling is already in progress")

        self._is_polling = True
        try:
            while True:
                # IdP-provided poll interval
                await sleep(codes.interval)

                r = await self._ctx.httpx_client.post(
                    f"{self._settings.domain}oauth/token",
                    auth=None,  # override client default
                    headers={"content-type": "application/x-www-form-urlencoded"},
                    data={
                        "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                        "device_code": codes.device_code,
                        "client_id": self._settings.client_id,
                    },
                )

                try:
                    resp: dict = r.json()
                except JSONDecodeError:
                    raise DeviceCodePollingError(
                        f"Unable to unpack IdP response: {r.content}"
                    )

                # Success!
                if r.status_code == 200:
                    self._validate_and_save_tokens(resp)

                    logger.debug(f"Got tokens: {self.tokens}")
                    return self

                # Keep polling
                try:
                    poll_err = PollingErrorResponse.model_validate(resp)
                except ValidationError as e:
                    raise DeviceCodePollingError(
                        f"Failed to unpack polling response: {r.text}"
                    ) from e
                if poll_err.error in [
                    PollingErrorType.AUTHORIZATION_PENDING,
                    PollingErrorType.SLOW_DOWN,
                ]:
                    continue

                # Timeout
                if poll_err.error == PollingErrorType.EXPIRED_TOKEN:
                    raise DeviceCodePollingExpiredError

                # Unauthorized
                if poll_err.error == PollingErrorType.ACCESS_DENIED:
                    raise UnauthorizedError

                # Unhandled
                if poll_err:
                    raise DeviceCodePollingError(f"Unknown polling error: {poll_err}")

        finally:
            self._is_polling = False

    async def _async_request_device_code(self, scope: Optional[str] = None):
        """
        Request new device code from IdP

        Args:
            scope: the specific oauth2 scopes to request

        Raises:
            DeviceCodeRequestDeviceCodeError: failed to get a device code from the IdP
        """
        scope = scope or self._settings.scope

        r = await self._ctx.httpx_client.post(
            f"{self._settings.domain}oauth/device/code",
            auth=None,  # override client default
            headers={"content-type": "application/x-www-form-urlencoded"},
            data={
                "client_id": self._settings.client_id,
                "scope": scope,
                "audience": self._settings.audience,
            },
        )

        if r.status_code != 200:
            raise DeviceCodeRequestDeviceCodeError(
                f"Failed to get a device code: {r.content}"
            )

        try:
            codes = GetDeviceCodeResponse.model_validate_json(r.content)
        except ValidationError as e:
            raise DeviceCodeRequestDeviceCodeError(
                f"Failed to unpack device code response: {r.text}"
            ) from e

        logger.debug(f"Got device code response: {codes}")
        return codes



import base64
import binascii
import datetime as dt
import fnmatch
import functools
from contextlib import suppress
from enum import Enum
from functools import cached_property
from typing import Annotated, Any, Optional, Type, Union

from annotated_types import Ge, Gt
from pydantic import (
    AliasChoices,
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    HttpUrl,
    ValidationError,
    model_validator,
)

from earthscope_sdk import __version__
from earthscope_sdk.model.secret import SecretStr


def _try_float(v: Any):
    try:
        return float(v)
    except (TypeError, ValueError):
        return v


Timedelta = Annotated[dt.timedelta, BeforeValidator(_try_float)]


class AccessTokenBody(BaseModel):
    """
    Access token payload

    [See Auth0 docs](https://auth0.com/docs/secure/tokens/access-tokens/access-token-profiles)
    """

    audience: Annotated[Union[str, list[str]], Field(alias="aud")]
    issuer: Annotated[str, Field(alias="iss")]
    issued_at: Annotated[dt.datetime, Field(alias="iat")]
    expires_at: Annotated[dt.datetime, Field(alias="exp")]
    scope: Annotated[str, Field(alias="scope")] = ""
    subject: Annotated[str, Field(alias="sub")]
    grant_type: Annotated[Optional[str], Field(alias="gty")] = None
    token_id: Annotated[Optional[str], Field(alias="jti")] = None
    client_id: Annotated[
        str,
        Field(
            validation_alias=AliasChoices("client_id", "azp"),
            serialization_alias="client_id",
        ),
    ]

    @cached_property
    def ttl(self) -> dt.timedelta:
        """time to live (TTL) until expiration"""
        return self.expires_at - dt.datetime.now(dt.timezone.utc)

    model_config = ConfigDict(extra="allow", frozen=True, populate_by_name=True)


class AuthFlowType(Enum):
    DeviceCode = "device_code"
    MachineToMachine = "m2m"


class Tokens(BaseModel):
    """
    EarthScope SDK oauth2 tokens
    """

    access_token: Optional[SecretStr] = None
    id_token: Optional[SecretStr] = None
    refresh_token: Optional[SecretStr] = None

    model_config = ConfigDict(frozen=True)

    @cached_property
    def access_token_body(self):
        if self.access_token is None:
            return None

        with suppress(IndexError, binascii.Error, ValidationError):
            payload_b64 = self.access_token.get_secret_value().split(".", 2)[1]
            payload = base64.b64decode(payload_b64 + "==")  # extra padding
            return AccessTokenBody.model_validate_json(payload)

        raise ValueError("Unable to decode access token body")

    @model_validator(mode="after")
    def ensure_one_of(self):
        # allow all fields to be optional in subclasses
        if self.__class__ != Tokens:
            return self

        if self.access_token or self.refresh_token:
            return self

        raise ValueError("At least one of access token and refresh token is required.")


class RetrySettings(BaseModel):
    """
    Retry configuration for the [Stamina library](https://stamina.hynek.me/en/stable/index.html)
    """

    # same defaults as AWS SDK "standard" mode:
    # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html#standard-retry-mode

    attempts: Annotated[int, Ge(0)] = 3
    timeout: Timedelta = dt.timedelta(seconds=20)

    wait_initial: Timedelta = dt.timedelta(milliseconds=100)
    wait_max: Timedelta = dt.timedelta(seconds=5)
    wait_jitter: Timedelta = dt.timedelta(seconds=1)
    wait_exp_base: Annotated[float, Gt(0)] = 2

    async def retry_context(self, *retry_exc: Type[Exception]):
        """
        Obtain a [Stamina](https://stamina.hynek.me/en/stable/index.html) retry iterator.
        """
        from stamina import retry_context

        retry_on = functools.partial(self.is_retriable, retry_exc=retry_exc)

        ctx = retry_context(
            on=retry_on,
            attempts=self.attempts,
            timeout=self.timeout,
            wait_initial=self.wait_initial,
            wait_jitter=self.wait_jitter,
            wait_max=self.wait_max,
            wait_exp_base=self.wait_exp_base,
        )
        async for attempt in ctx:
            yield attempt

    def is_retriable(
        self,
        exc: Exception,
        *args,
        retry_exc: tuple[Type[Exception]] = (),
        **kwargs,
    ) -> bool:
        """
        Check if the given exception can be retried
        """
        if retry_exc and isinstance(exc, retry_exc):
            return True

        return False


class HttpRetrySettings(RetrySettings):
    status_codes: set[int] = {429, 500, 502, 503, 504}

    def is_retriable(
        self,
        exc: Exception,
        *args,
        **kwargs,
    ) -> bool:
        from httpx import HTTPStatusError

        if isinstance(exc, HTTPStatusError):
            if exc.response.status_code in self.status_codes:
                return True

        return super().is_retriable(exc, *args, **kwargs)


class AuthFlowSettings(Tokens):
    """
    Auth flow configuration

    Not for direct use.
    """

    # Auth parameters
    audience: str = "https://api.earthscope.org"
    domain: HttpUrl = HttpUrl("https://login.earthscope.org")
    client_id: str = "b9DtAFBd6QvMg761vI3YhYquNZbJX5G0"
    scope: str = "offline_access"
    client_secret: Optional[SecretStr] = None

    # Only inject bearer token for requests to these hosts
    allowed_hosts: set[str] = {
        "earthscope.org",
        "*.earthscope.org",
    }

    # Auth exchange retries
    retry: HttpRetrySettings = HttpRetrySettings(
        attempts=5,
        timeout=dt.timedelta(seconds=30),
        wait_initial=dt.timedelta(seconds=1),
        wait_jitter=dt.timedelta(seconds=3),
    )

    @cached_property
    def auth_flow_type(self) -> AuthFlowType:
        if self.client_secret is not None:
            return AuthFlowType.MachineToMachine

        return AuthFlowType.DeviceCode

    @cached_property
    def allowed_host_patterns(self) -> set[str]:
        """
        The subset of allowed hosts that are glob patterns.

        Use `is_host_allowed` to check if a host is allowed by any of these patterns.
        """
        return {h for h in self.allowed_hosts if "*" in h or "?" in h}

    def is_host_allowed(self, host: str) -> bool:
        """
        Check if a host matches any pattern in the allowed hosts set.

        Supports glob patterns with '?' and '*' characters (e.g., *.earthscope.org).

        Args:
            host: The hostname to check

        Returns:
            True if the host matches any allowed pattern, False otherwise
        """
        if host in self.allowed_hosts:
            return True

        for allowed_pattern in self.allowed_host_patterns:
            if fnmatch.fnmatch(host, allowed_pattern):
                self.allowed_hosts.add(host)
                return True

        return False


class HttpSettings(BaseModel):
    """
    HTTP client configuration
    """

    # httpx limits
    keepalive_expiry: Timedelta = dt.timedelta(seconds=5)
    max_connections: Optional[int] = None
    max_keepalive_connections: Optional[int] = None

    # httpx timeouts
    timeout_connect: Timedelta = dt.timedelta(seconds=5)
    timeout_read: Timedelta = dt.timedelta(seconds=5)

    # automatically retry requests
    retry: HttpRetrySettings = HttpRetrySettings()

    # Other
    user_agent: str = f"earthscope-sdk py/{__version__}"

    @cached_property
    def limits(self):
        """httpx Limits on client connection pool"""
        # lazy import
        import httpx

        return httpx.Limits(
            max_connections=self.max_connections,
            max_keepalive_connections=self.max_keepalive_connections,
            keepalive_expiry=self.keepalive_expiry.total_seconds(),
        )

    @cached_property
    def timeouts(self):
        """httpx Timeouts default behavior"""
        # lazy import
        import httpx

        return httpx.Timeout(
            connect=self.timeout_connect.total_seconds(),
            # reuse read timeout for others
            read=self.timeout_read.total_seconds(),
            write=self.timeout_read.total_seconds(),
            pool=self.timeout_read.total_seconds(),
        )


class ResourceRefs(BaseModel):
    """
    References to EarthScope resources
    """

    api_url: HttpUrl = HttpUrl("https://api.earthscope.org")
    """Base URL for api.earthscope.org"""


class SdkBaseSettings(BaseModel):
    """
    Common base class for SDK settings

    Not for direct use.
    """

    http: HttpSettings = HttpSettings()
    oauth2: AuthFlowSettings = AuthFlowSettings()
    resources: ResourceRefs = ResourceRefs()



from contextlib import suppress
from datetime import timedelta
from typing import TYPE_CHECKING, Optional

from earthscope_sdk.client.user._base import UserBaseService

if TYPE_CHECKING:
    from earthscope_sdk.client.user.models import AwsTemporaryCredentials
    from earthscope_sdk.common.context import SdkContext


class _UserService(UserBaseService):
    """
    L2 user/IDM service functionality
    """

    def __init__(self, ctx: "SdkContext"):
        super().__init__(ctx)

        self._aws_creds_by_role = dict[str, "AwsTemporaryCredentials"]()

    async def _get_aws_credentials(
        self,
        *,
        role: str = "s3-miniseed",
        force=False,
        ttl_threshold: Optional[timedelta] = None,
    ):
        """
        Retrieve temporary AWS credentials.

        Leverages a memory cache and disk cache in this profile's state directory
        (`~/.earthscope/<profile_name>/aws.<role>.json`)

        Args:
            role: Alias of the role to assume. Defaults to `s3-miniseed`.
            force: Ignore cache and fetch new credentials. Defaults to `False`.
            ttl_threshold: Time-to-live remaining on cached credentials after which new credentials are fetched. Defaults to 30s.

        Returns: Temporary AWS credentials for the selected role.
        """
        # lazy imports
        from pydantic import ValidationError

        from earthscope_sdk.client.user.models import AwsTemporaryCredentials

        aws_creds_file_path = self.ctx.settings.profile_dir / f"aws.{role}.json"

        if not force:
            # Check memory cache
            if aws_creds := self._aws_creds_by_role.get(role):
                if not aws_creds.is_expired(ttl_threshold):
                    return aws_creds

            # Check disk cache
            with suppress(FileNotFoundError, ValidationError):
                aws_creds_bytes = aws_creds_file_path.read_bytes()
                aws_creds = AwsTemporaryCredentials.model_validate_json(aws_creds_bytes)
                if not aws_creds.is_expired(ttl_threshold):
                    return aws_creds

        # Get new AWS creds
        # NOTE: this method will implicitly refresh our access token if necessary
        aws_creds = await super()._get_aws_credentials(role=role)

        # persist in memory and on disk
        self._aws_creds_by_role[role] = aws_creds
        aws_creds_file_path.write_bytes(aws_creds.model_dump_json().encode())

        return aws_creds


class AsyncUserService(_UserService):
    """
    User and Identity Management functionality
    """

    def __init__(self, ctx: "SdkContext"):
        super().__init__(ctx)

        self.get_aws_credentials = self._get_aws_credentials
        self.get_profile = self._get_profile


class UserService(_UserService):
    """
    User and Identity Management functionality
    """

    def __init__(self, ctx: "SdkContext"):
        super().__init__(ctx)

        self.get_aws_credentials = ctx.syncify(self._get_aws_credentials)
        self.get_profile = ctx.syncify(self._get_profile)



from functools import cached_property

from earthscope_sdk.common.client import AsyncSdkClient, SdkClient


class AsyncEarthScopeClient(AsyncSdkClient):
    """
    An async client for interacting with api.earthscope.org
    """

    @cached_property
    def user(self):
        """
        User and Identity Management functionality
        """
        # lazy load
        from earthscope_sdk.client.user._service import AsyncUserService

        return AsyncUserService(self._ctx)


class EarthScopeClient(SdkClient):
    """
    A client for interacting with api.earthscope.org
    """

    @cached_property
    def user(self):
        """
        User and Identity Management functionality
        """
        # lazy load
        from earthscope_sdk.client.user._service import UserService

        return UserService(self._ctx)
